/********************************************************************
 *	 Copyright (c) 2006, Joachim Zobel
 *	 Author: Joachim Zobel <jz-2006@heute-morgen.de>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*********************************************************************/

/*
 	mod_delay
	is an output filter that does buffering until it is told to stop
	this (by a note). It also enables the sending of 304 by delaying the
	call to ap_meets_conditions until then.

	The module exports one optional functions and has no configuration.
	It registers as an output filter named delay.
 */

#include <ctype.h>

#include <http_protocol.h>
#include <http_config.h>
#include <util_filter.h>
#include <http_log.h>

#include "mod_delay.h"


typedef struct delay_ctx {
  apr_bucket_brigade* bb ;
  /* store flag from request */
  int no_local_copy ;
} delay_ctx ;

typedef int delay_flag;

/*+
 * @var AP_NOTE_DELAY Identifies the note that this module uses 
 * for communication between filters.
 */
#define	NOT_INITIALIZED	(-1)
static apr_size_t AP_NOTE_DELAY	= NOT_INITIALIZED ;

/*+
 * Get the delay flag
 * @param r The request
 * @return The delay flag (0, 1)
 */
static delay_flag delay_get_flag(request_rec *r)
{
  delay_flag *p_delay=  *((delay_flag **)ap_get_request_note(r, AP_NOTE_DELAY)) ;
  return *p_delay ;
}

/*+
 * Ends delaying by setting the delay flag to 0
 * @param r The request
 */
static void delay_end(request_rec *r)
{
  delay_flag *p_delay=  NULL;
  if (AP_NOTE_DELAY != NOT_INITIALIZED) {
    p_delay 	= *((delay_flag **)ap_get_request_note(r, AP_NOTE_DELAY)) ;
    *p_delay	= 0 ;
  }
}

/*+
 * Initializes the delay flag with 1.
 * @param r The request
 */
static void delay_flag_init(request_rec *r)
{
  delay_flag *p_delay	= apr_palloc(r->pool, sizeof(delay_flag)) ;
  *p_delay	= 1 ;
		
  if (AP_NOTE_DELAY == NOT_INITIALIZED) {
	AP_NOTE_DELAY = ap_register_request_note() ;
  }
	
  *((delay_flag **)ap_get_request_note(r, AP_NOTE_DELAY))  = p_delay;  
}

/*+
 * Initializes the delay filter
 * @param f The filter
 * @return OK 
 */
static int delay_filter_init(ap_filter_t *f)
{
  delay_ctx	* fctx ;
  
  fctx	 = f->ctx = apr_pcalloc(f->r->pool, sizeof(delay_ctx)) ;
  
  fctx->bb	= NULL ;
  /* save nlc flag */
  fctx->no_local_copy	= f->r->no_local_copy ;
  f->r->no_local_copy	= 1 ;
  
  delay_flag_init(f->r) ;
  
  return OK;
}

/*+
 * The delay filtering function
 * @param f The filter
 * @param bb This runs bucket brigade
 * @return  
 */
static int delay_filter(ap_filter_t* f, apr_bucket_brigade* bb) {
  int rtn = APR_SUCCESS ;
  delay_ctx* ctxt = (delay_ctx*)f->ctx ;
  apr_bucket* b ;
  int last = 0 ;

  /* restore nlc flag */
  f->r->no_local_copy = ctxt->no_local_copy	;

  /* Is this the last brigade ? */
  for ( b = APR_BRIGADE_FIRST(bb) ;
	b != APR_BRIGADE_SENTINEL(bb) ;
	b = APR_BUCKET_NEXT(b) ) {
    if ( APR_BUCKET_IS_EOS(b) ) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "Found eos bucket.");
      last = 1;
    }
  }
  
  if ( ! ctxt )
    return ap_pass_brigade(f->next, bb) ;

  if (!last && delay_get_flag(f->r)) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "Delaying brigade.");
    /* 
     * As long as delay is turned on we store all brigades
     * in the filter context
     */
    if (ctxt->bb) {
      APR_BRIGADE_CONCAT(ctxt->bb, bb) ;
    }
    else {
     ctxt->bb	= bb ;
    }    
  }
  /**
   * This is _not_ an else since the delay
   * flag may have changed during the execution of
   * the above code.
   */
  if (last || !delay_get_flag(f->r)) {
    rtn = ap_meets_conditions(f->r) ;
    if ( rtn != OK ) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "Exiting filter with %d.", rtn);
      f->r->status = rtn ;
      return APR_SUCCESS ;
    }
    if (ctxt->bb) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "Sending stored brigade.");
      /*
       * We send out the stored brigades
       */
      APR_BRIGADE_CONCAT(ctxt->bb, bb) ;
      ctxt->bb	= NULL ;
    }
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r, "Passing brigade.");
    }

    rtn = ap_pass_brigade(f->next, bb) ;
  }
  
  return rtn ;
}

static void delay_hooks(apr_pool_t* p) {
  ap_register_output_filter("delay", delay_filter,
	delay_filter_init, AP_FTYPE_RESOURCE) ;
  APR_REGISTER_OPTIONAL_FN(delay_end) ;
}

module AP_MODULE_DECLARE_DATA delay_module = {
	STANDARD20_MODULE_STUFF,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	delay_hooks
} ;


